@@ -62,6 +62,10 @@ from ..django import get_list
 from ..django import get_key_from_request_data_or_self_obj
 from ..django.utils import get_field_type_in_db, get_field_type_in_py
 from bdtime.datetime_utils import common_date_time_formats
+from ..django.utils import get_field_names_by_model
+import sys
+from django.http import JsonResponse
+from .admin_env_init import DATA_UPLOAD_MAX_NUMBER_FIELDS
 
 
 IAdmin = admin.ModelAdmin
@@ -350,7 +354,7 @@ class ExportCsvMixin:
     export_as_csv.icon = 'fas fa-download'
 
 
-class ImportAdmin(IDAdmin):
+class ImportMixin:
     """
     导入类, CSV和Excel通用
 
@@ -366,6 +370,16 @@ class ImportAdmin(IDAdmin):
 
     use_bulk_create = False     # 是否使用`bulk_create`加快新建数据的速度, 但同时也会不调用`model.save()`方法
     export_all_fields_forever = False       # 是否一直导出所有字段(右上角的`导出全部`和`action`中的导出数据做区分)
+    # raise_error_to_debug = True     # debug模式
+    raise_error_to_debug = False     # debug模式
+
+    bulk_size = 50000       # 批处理数量
+    # bulk_size = 3000       # 批处理数量
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.CHECK_EXIST_FLAG = False
+        self.CREATE_EXIST_FLAG = False
 
     def import_csv(self, request):
         t_import = Time()
@@ -402,13 +416,21 @@ class ImportAdmin(IDAdmin):
 
                 verbose_names = list(field_dc.keys())
 
+                # 判断哪些需要在pd读取时指定为text字段
                 dtype = None
                 for i in range(len(field_names)):
                     field_name = field_names[i]
                     field_type = get_field_type_in_py(self.model, field_name)
                     verbose_name = verbose_names[i]
+                    if field_type in ['ForeignKey', 'OneToOneField', 'ManyToManyField']:
+                        print(field_name, field_type, verbose_name)
+                        _meta_field = meta.fields[i]
+                        related_model = _meta_field.related_model
+                        to_field = _meta_field.to_fields[0]
+                        if get_field_type_in_py(related_model, to_field) == 'str':
+                            field_type = 'str'
+
                     if field_type == 'str':
-                        print(f'{field_name} --- {verbose_name} --- {field_type}')
                         dc = {
                             verbose_name: str,
                         }
@@ -446,11 +468,23 @@ class ImportAdmin(IDAdmin):
                 else:
                     raise TypeError(f'文件格式错误! 仅支持{format_ls}格式的文档.')
 
+                df = df.dropna(axis=0, how='all', thresh=None, subset=None, inplace=False)  # 只有全部为空才回被删除
                 df_rows = df.shape[0]        # 一共多少行数据
+                assert df_rows <= DATA_UPLOAD_MAX_NUMBER_FIELDS, \
+                    f'检测到{df_rows}条数据, 超过了最大上传数量{DATA_UPLOAD_MAX_NUMBER_FIELDS}条!'
 
                 # 删除Unnamed列
                 df = df.loc[:, ~df.columns.str.match('Unnamed')]
 
+                pk_i = None
+                for i in range(len(meta.fields)):
+                    _f = meta.fields[i]
+                    if _f.primary_key:
+                        pk_i = i
+                        break
+                pk_name = field_names[pk_i]
+                # pk_type = get_field_type_in_py(self.model, pk_name)
+
                 valid_columns = [column_i in field_names or column_i in verbose_names for column_i in df.columns]
                 df = df.loc[:, valid_columns]
                 titles = df.columns.tolist()
@@ -460,98 +494,229 @@ class ImportAdmin(IDAdmin):
                 df1 = df.copy()
                 df1.columns = title_ls
 
+                # title到verbose_name的转换字典
+                verbose_ls = list(df.columns)
+                title_to_verbose_dc = dict(zip(title_ls, verbose_ls))
+                # title_to_verbose_dc.update({
+                #     'pk': 'ID',
+                #     'id': 'ID',
+                # })
+                info_dc = {
+                    'title_to_verbose_dc': title_to_verbose_dc,
+                }
+
+                # 主键去重
+                if pk_name in df1.columns:
+                    col_pk: pd.Series = df1[pk_name]
+                    duplicated_ls = np.where(col_pk.duplicated())[0]
+                    duplicated_length = len(duplicated_ls)
+                    if duplicated_length:
+                        msg = f'主键id有重复!'
+                        if duplicated_length < 10:
+                            msg += f'重复位置: 第{list((duplicated_ls + 2))}行'
+                        else:
+                            msg += f'重复次数: {duplicated_length}次, 重复位置: 第{list((duplicated_ls[:5] + 2))}行'
+
+                        assert len(duplicated_ls) == 0, msg
+
                 # 这里有问题, 服务器带宽太小的话, 一次性bulk_update会造成tcp拥塞, 从而导致服务器瘫痪!
                 # 必须仿照navicat的处理方式, 设置batch_size, 小批量处理数据.
                 # 同时出现问题, 如何让用户看到进度条? 用异步的方式, 或者websocket?
-                md_ls = []
                 md_dc_ls = []
                 my_tqdm = tqdm(total=df_rows)
-                for index, row in df1.iterrows():
-                    content_ls = row.values.tolist()
+                my_tqdm.desc = '数据校验中...'
+                import queue
+                q = queue.Queue()
 
-                    # 处理DateField字段 and 外键 and 主键
-                    for i in range(len(title_ls)):
-                        title_i = title_ls[i]
-                        content_ls[i] = conv_nan(content_ls[i])
+                def check_df_format(self, df, md_dc_ls, curr_id, start=0, end=None, n=None, tqdm_i=None, info_dc=info_dc, q=q):
+                    from bdtime import show_json, show_ls
 
-                        attr = getattr(self.model, title_i)
-                        if not hasattr(attr, 'field'):
-                            continue
+                    n = len(df) if n is None else n
+                    if not n:
+                        return
 
-                        attr_field_name = attr.field.__class__.__name__
+                    if end is None:
+                        end = n
+                    if end > n:
+                        end = n
 
-                        if attr_field_name in ['ForeignKey', 'OneToOneField', 'ManyToManyField']:
-                            # print(f'处理外键字段 --- {title_i} --- {content_ls[i]}')
-                            _meta_field = meta.fields[i]
-                            from bdtime import show_json, show_ls
-                            # show_json(_meta_field.__dict__)
-                            related_model = _meta_field.related_model
-                            to_field = _meta_field.to_fields[0]
-                            to_field = to_field if to_field else 'pk'
-                            field_value = content_ls[i]
-                            _obj_qs_ls = related_model.objects.filter(**{to_field: field_value})
-                            assert _obj_qs_ls.exists(), f'第[{i}]行外键[{title_i}]列找不到对应[{related_model._meta.verbose_name}_{to_field}={field_value}]的值!'
-                            _obj = _obj_qs_ls[0]
-                            content_ls[i] = _obj
+                    title_to_verbose_dc = info_dc.get('title_to_verbose_dc')
 
-                        if attr_field_name in ['DateField', 'DateTimeField']:
-                            if attr_field_name == 'DateTimeField':
-                                res = conv_date_time_field_str_format(ts=content_ls[i])
-                                # if res is None:
-                                #     print(content_ls[i])
-                                content_ls[i] = res
-                            else:
-                                ts = content_ls[i]
-                                res = conv_date_field_str_format(ts=ts) or conv_date_time_field_str_format(ts=content_ls[i])
-                                if res:
+                    _df = df[start: end]
+
+                    try:
+                        for index, row in _df.iterrows():
+                            # print('CHECK_EXIST_FLAG --- ', self.CHECK_EXIST_FLAG)
+                            if self.CHECK_EXIST_FLAG:
+                                return
+
+                            content_ls = row.values.tolist()
+
+                            # 处理DateField字段 and 外键 and 主键
+                            for i in range(len(title_ls)):
+                                title_i = title_ls[i]
+                                content_ls[i] = conv_nan(content_ls[i])
+                                verbose_i = title_to_verbose_dc.get(title_i)
+                                # if title_i in ['id', 'ID', 'pk', 'PK']:
+                                title_type_i = get_field_type_in_py(self.model, title_i)
+
+                                attr = getattr(self.model, title_i)
+                                if not hasattr(attr, 'field'):
+                                    continue
+                                attr_field_name = attr.field.__class__.__name__
+                                field_index = get_field_names_by_model(self.model).index(title_i)
+                                if attr_field_name in ['ForeignKey', 'OneToOneField', 'ManyToManyField']:
+                                    # print(f'处理外键字段 --- {title_i} --- {content_ls[i]}')
+                                    _meta_field = meta.fields[field_index]
+                                    related_model = _meta_field.related_model
+                                    to_field = _meta_field.to_fields[0]
+                                    to_field = to_field if to_field else 'pk'
+                                    field_value = content_ls[i]
+                                    _obj_qs_ls = related_model.objects.filter(**{to_field: field_value})
+                                    assert _obj_qs_ls.exists(), f'第[{index + 2}]行外键[{verbose_i}]列找不到对应[{related_model._meta.verbose_name}_{to_field}={field_value}]的值!'
+                                    _obj = _obj_qs_ls[0]
+                                    content_ls[i] = _obj
+                                elif attr_field_name in ['DateField', 'DateTimeField']:
+                                    if attr_field_name == 'DateTimeField':
+                                        res = conv_date_time_field_str_format(ts=content_ls[i])
+                                        content_ls[i] = res
+                                    else:
+                                        ts = content_ls[i]
+                                        res = conv_date_field_str_format(ts=ts) or conv_date_time_field_str_format(
+                                            ts=content_ls[i])
+                                        if res:
+                                            try:
+                                                if '.' in res:
+                                                    res_dt = datetime.datetime.strptime(res,
+                                                                                        common_date_time_formats.ms_dt)
+                                                elif ':' in res:
+                                                    res_dt = datetime.datetime.strptime(res,
+                                                                                        common_date_time_formats.s_dt)
+                                                else:
+                                                    res_dt = datetime.datetime.strptime(res,
+                                                                                        common_date_time_formats.only_date)
+                                                res = res_dt.strftime(common_date_time_formats.only_date)
+                                            except Exception as e:
+                                                raise ValueError(
+                                                    f'字段[{title_i}]日期格式错误! 标准格式: [%Y-%m-%d]. error: {e}')
+                                        content_ls[i] = res
+                                elif title_type_i == 'int':
                                     try:
-                                        if '.' in res:
-                                            res_dt = datetime.datetime.strptime(res, common_date_time_formats.ms_dt)
-                                        elif ':' in res:
-                                            res_dt = datetime.datetime.strptime(res, common_date_time_formats.s_dt)
-                                        else:
-                                            res_dt = datetime.datetime.strptime(res, common_date_time_formats.only_date)
-                                        res = res_dt.strftime(common_date_time_formats.only_date)
-                                    except Exception as e:
-                                        raise ValueError(f'字段[{title_i}]日期格式错误! 标准格式: [%Y-%m-%d]. error: {e}')
-                                content_ls[i] = res
+                                        if content_ls[i]:
+                                            int(content_ls[i])
+                                    except Exception as _:
+                                        raise TypeError(f'第[{index + 2}]行[{verbose_i}]无法转换为`整数`类型! 值为: [{content_ls[i]}]')
+                                elif title_type_i == 'float':
+                                    try:
+                                        if content_ls[i]:
+                                            float(content_ls[i])
+                                    except Exception as _:
+                                        raise TypeError(f'第[{index + 2}]行[{verbose_i}]无法转换为`小数`类型! 值为: [{content_ls[i]}]')
+                                elif title_type_i == 'bool':
+                                    try:
+                                        if content_ls[i]:
+                                            bool(content_ls[i])
+                                    except Exception as _:
+                                        raise TypeError(f'第[{index + 2}]行[{verbose_i}]无法转换为`bool`类型! 值为: [{content_ls[i]}]')
 
-                    dc = dict(zip(title_ls, content_ls))
+                            dc = dict(zip(title_ls, content_ls))
 
-                    if not (self.use_original_id and set(df1.columns).intersection(set(['id', 'ID', 'pk', 'PK']))):
-                        # 如果[不使用原始id 并且 表格中有`id`列],  则启用自定义id
-                        dc.update({'pk': curr_id})
-                        curr_id += 1
+                            if not (self.use_original_id and set(_df.columns).intersection(
+                                    set(['id', 'ID', 'pk', 'PK']))):
+                                # 如果[不使用原始id 并且 表格中有`id`列],  则启用自定义id
+                                dc.update({'pk': curr_id})
+                                curr_id += 1
 
-                    if self.use_bulk_create:
-                        md = self.model(**dc)
-                        md_ls.append(md)
-                    else:
-                        md_dc_ls.append(dc)
+                            md_dc_ls.append(dc)
+                            my_tqdm.update(1)
 
-                    my_tqdm.update(1)
+                        if tqdm_i:
+                            tqdm_i.update(1)
+                    except Exception as e:
+                        print('\n\n ****** check error --------', e)
+                        if q is not None:
+                            self.CHECK_EXIST_FLAG = True
+                            q.put(sys.exc_info())
+                        else:
+                            raise e
 
-                    # self.model.objects.create(**dc)
-
-                    # if self.use_bulk_create is False:
-                    #     self.model.objects.create(**dc)
-
-                if self.use_bulk_create:
-                    self.model.objects.bulk_create(md_ls, batch_size=100)
+                n = df_rows
+                bulk_size = self.bulk_size
+                if n < bulk_size:
+                    check_df_format(self, df1, md_dc_ls, curr_id, start=0, end=df_rows, n=df_rows, tqdm_i=my_tqdm, info_dc=info_dc, q=None)
                 else:
-                    save_tqdm = tqdm(total=df_rows)
-                    save_tqdm.desc = '写入数据库ing...'
-                    for md_dc_i in md_dc_ls:
+                    a = np.arange(n)
+                    s = slice(0, len(a), bulk_size)
+                    start_end_ls = list(a[s])
+                    thread_ls = []
+                    for start_end_i in start_end_ls:
+                        thread_i = threading.Thread(target=check_df_format,
+                                                    args=(self, df1, md_dc_ls, curr_id, start_end_i, start_end_i + bulk_size, df_rows, my_tqdm, info_dc, q))
+                        thread_ls.append(thread_i)
+                        thread_i.start()
+
+                    self.check_thread_ls(thread_ls, q, 'CHECK_EXIST_FLAG')
+
+                save_tqdm = tqdm(total=df_rows)
+                save_tqdm.desc = '写入数据库ing...'
+                print(f'*** df_rows: {df_rows} --- len(md_dc_ls): {len(md_dc_ls)}')
+                save_q = queue.Queue()
+
+                def save_md_dc_ls(md_dc_ls, start=0, end=None, n=None, tqdm_i=None, q=None):
+                    n = len(md_dc_ls) if n is None else n
+                    if not n:
+                        return
+
+                    if end is None:
+                        end = n
+                    if end > n:
+                        end = n
+                    try:
+                        for i in range(start, end):
+                            if self.CREATE_EXIST_FLAG:
+                                return
+                            md_dc_i = md_dc_ls[i]
+                            self.model.objects.create(**md_dc_i)  # `md_i.save()`不太好, 容易覆盖原有数据
+                            if tqdm_i:
+                                tqdm_i.update(1)
+                    except Exception as e:
+                        if q is not None:
+                            q.put(sys.exc_info())
+                        else:
+                            raise e
+
+                from ..django.conf import db_engine
+                use_sqlite3 = 'sqlite3' in db_engine
+                if n < bulk_size or use_sqlite3:
+                    for i in range(len(md_dc_ls)):
+                        md_dc_i = md_dc_ls[i]
                         save_tqdm.update(1)
                         self.model.objects.create(**md_dc_i)                 # `md_i.save()`不太好, 容易覆盖原有数据
+                    msg = f"{f_format}文件导入成功! 一共导入{df_rows}条数据, 耗时: {t_import.now(1)}秒."
+                else:
+                    a = np.arange(n)
+                    s = slice(0, len(a), bulk_size)
+                    start_end_ls = list(a[s])
+                    thread_ls = []
+                    for start_end_i in start_end_ls:
+                        thread_i = threading.Thread(target=save_md_dc_ls, args=(md_dc_ls, start_end_i, start_end_i + bulk_size, n, save_tqdm, save_q))
+                        thread_ls.append(thread_i)
+                        thread_i.start()
+
+                    # self.CREATE_EXIST_FLAG = False
+                    self.check_thread_ls(thread_ls, save_q, 'CREATE_EXIST_FLAG')
+
+                    # for thread_i in thread_ls:
+                    #     thread_i.join()
+                    msg = f"{f_format}文件导入中...<br>一共将导入{df_rows}条数据, <br>请稍后刷新查看是否完全导入."
 
                 reset_db_sequence(self.model)
-                self.message_user(request, f"{f_format}文件导入成功! 一共导入{df_rows}条数据, 耗时: {t_import.now(1)}秒.")
+                self.message_user(request, msg)
                 self.remove_temp_file(tempdir)
 
                 return redirect("..")
         except Exception as e:
-            msg = f"第 {index+1} 条数据导入失败!</br>错误信息：&nbsp;" + str(e)
+            msg = f"数据导入失败!</br>错误信息：&nbsp;" + str(e)
 
             ret_restful_api = request._post.get('ret_restful_api')  # 是否返回RestfulAPI格式
 
@@ -560,20 +725,50 @@ class ImportAdmin(IDAdmin):
             if not convert_query_parameter_to_bool(ret_restful_api):
                 self.message_user(request, msg, level=messages.ERROR)
             else:
-                # from ..django.utils import my_api_assert_function, APIResponse
-                # my_api_assert_function(False, msg)
-                from django.http import JsonResponse
-                # return APIResponse(ret=None, msg=msg, status=404)
                 return JsonResponse(data={
                     'status': 'error',
                     'msg': msg
                 })
+
+            _msg = msg
+            _msg = _msg.replace('&nbsp;', ' ')
+            _msg = _msg.replace('<br>', '\n')
+            _msg = _msg.replace('</br>', '\n')
+            if self.raise_error_to_debug:
+                raise ImportError(f'\n\n ****** Import Data Error:\n\n{_msg}\n\n******************\n\n')
+            else:
+                print(f'\n\n ****** Import Data Error:\n\n{_msg}\n\n******************\n\n')
             return redirect("..")
 
         form = CsvImportForm()
         payload = {"form": form}
         return render(request, "admin/csv_form.html", payload)
 
+    def check_thread_ls(self, thread_ls, q, flag_attr_name):
+        from bdtime import tt
+        while True:
+            tt.sleep(0.1)
+            if q.empty():
+                count = 0
+                for thread_i in thread_ls:
+                    if thread_i.is_alive() == False:
+                        count += 1
+                if count == len(thread_ls):
+                    break
+                # if threading.active_count() == 1:
+                #     break
+            else:
+                last = q.get()
+
+                assert hasattr(self, flag_attr_name), f'check_thread_ls 不存在属性名[{flag_attr_name}]?'
+                setattr(self, flag_attr_name, True)
+
+                try:
+                    raise Exception(f'{last[-2]}' if len(last) >= 2 else f'{last[-1]}')
+                except:
+                    raise Exception(last)
+        print(u'end')
+
     def bulk_save(self, request, md_ls, f_format, df_rows, t_import, tempdir, batch_size=100):
         def _bulk_save():
             self.model.objects.bulk_create(md_ls, batch_size=batch_size)
@@ -647,6 +842,10 @@ class ImportAdmin(IDAdmin):
         return ret
 
 
+class ImportAdmin(ImportMixin, IDAdmin):
+    pass
+
+
 class CsvImportExportAdmin(ImportAdmin, ExportCsvMixin):
     """
     CSV导入/导出Admin类
