@@ -10,6 +10,7 @@ import threading
 import shutil
 from bdtime import Time
 import datetime as dt
+import sys
 
 
 TEMPDIR = 'tempdir'     # 临时文件夹
@@ -283,3 +284,12 @@ class SetUtils:
 
 set_utils = SetUtils()
 
+
+def replace_path_by_platform(path):
+    assert isinstance(path, str), 'path必须为str类型!'
+    if sys.platform == 'win32':
+        path = path.replace('/', '\\')
+    else:
+        path = path.replace('\\', '/')
+    return path
+
